﻿/**
 * Papervision3D Gallery Tutorial Example by FlashMoto.com
 *
 * @author Demetrio Fortman
 * @langversion ActionScript 3.0
 * @playerversion Flash 9.0
 *
 * http://www.flashmoto.com
 * demetrio.fortman@gmail.com
 *
 */
package com.flashmoto.gallery
{
	import br.com.stimuli.loading.BulkErrorEvent;
	import br.com.stimuli.loading.BulkLoader;
	import br.com.stimuli.loading.BulkProgressEvent;

	import caurina.transitions.Tweener;

	import flash.display.Bitmap;
	import flash.display.MovieClip;
	import flash.display.Sprite;
	import flash.events.Event;
	import flash.geom.Rectangle;
	import flash.system.Capabilities;
	import flash.text.TextField;

	import org.papervision3d.Papervision3D;
	import org.papervision3d.cameras.Camera3D;
	import org.papervision3d.core.log.PaperLogger;
	import org.papervision3d.core.log.PaperTraceLogger;
	import org.papervision3d.events.InteractiveScene3DEvent;
	import org.papervision3d.materials.BitmapMaterial;
	import org.papervision3d.materials.MovieMaterial;
	import org.papervision3d.objects.DisplayObject3D;
	import org.papervision3d.objects.primitives.Plane;
	import org.papervision3d.render.BasicRenderEngine;
	import org.papervision3d.scenes.Scene3D;
	import org.papervision3d.view.Viewport3D;

	public class Gallery extends Sprite
	{
		/**
		 * Loader for the gallery xml and gallery images
		 */
		public var loader:BulkLoader;

		/**
		 * Papervision3d components - Viewport
		 */
		public var viewport:Viewport3D;

		/**
		 * Papervision3d components - Scene
		 */
		public var scene:Scene3D;

		/**
		 * Papervision3d components - Camera
		 */
		public var camera:Camera3D;

		/**
		 * Papervision3d components - BasicRenderEngine
		 */
		public var renderer:BasicRenderEngine;

		/**
		 * Collection of images urls
		 */
		public var galleryImagesCollection:Array;

		/**
		 * Collection of gallery planes to be added to the scene
		 */
		public var galleryPlanesCollection:Array;

		/**
		 * Collection of scene gallery planes
		 */
		public var galleryPlanes:Array;

		/**
		 * Flag that indicates if xml is loaded
		 */
		private var _galleryXMLLoaded:Boolean = false;

		/**
		 * Holder for the gallery planes
		 */
		private var _planesHolder:DisplayObject3D;

		/**
		 * Flag that indicates zoomed image state
		 */
		private var _imageZoomed:Boolean = false;

		/**
		 * Current plane ( first in the line ) index
		 */
		private var _currentPlaneIndex:uint = 0;

		/**
		 * First plane z coordinate, planes shift
		 */
		private var _firstPlaneZ:Number = 0;

		/**
		 * Flag that indicates if planes are moving
		 */
		private var _planesAreMoving:Boolean = false;

		/**
		 * Path to the gallery xml file
		 */
		public static const GALLERY_XML:String = 'xml/gallery.xml';

		/**
		 * Viewport width
		 */
		public static const VIEWPORT_WIDTH:Number = 800;

		/**
		 * Viewport height
		 */
		public static const VIEWPORT_HEIGHT:Number = 800;

		/**
		 * Gallery plane width
		 */
		public static const GALLERY_PLANE_WIDTH:Number = 400;

		/**
		 * Gallery plane height
		 */
		public static const GALLERY_PLANE_HEIGHT:Number = 400;

		/**
		 * Gallery plane segments ( width )
		 */
		public static const GALLERY_PLANE_SEGMENTS_W:Number = 3;

		/**
		 * Gallery plane segments ( height )
		 */
		public static const GALLERY_PLANE_SEGMENTS_H:Number = 3;

		/**
		 * Gallery Constructor
		 */
		public function Gallery()
		{
			// Load images data
			loadImagesData();
		}

		/**
		 * Loads images data from xml file.
		 */
		private function loadImagesData():void
		{
			// Create new loader
			loader = new BulkLoader( 'galleryLoader' );
			// Loader events
			loader.addEventListener(BulkLoader.COMPLETE, onGalleryLoaded);
			loader.addEventListener(BulkLoader.PROGRESS,
				onGalleryLoadingProgress);
			loader.addEventListener(BulkErrorEvent.ERROR,
				onGalleryLoadingError);
			// Add gallery xml to the quenue
			loader.add( GALLERY_XML );
			// Start loading
			loader.start( 1 );
		}

		/**
		 * Parses gallery xml file and starts images loading process.
		 */
		private function onGalleryLoaded( event:Event ):void
		{
			if( !_galleryXMLLoaded ) {
				// Parse loaded gallery xml
				galleryImagesCollection = new Array();
				var galleryImagesData:XML = loader.getXML( GALLERY_XML, true );
				for each ( var image:XML in galleryImagesData.elements() ) {
					var imageSrc:String = image.@src.toString();
					if( imageSrc && imageSrc.length > 0 ) {
						galleryImagesCollection.push( imageSrc );
						loader.add( imageSrc );
					}
				}
				_galleryXMLLoaded = true;
				// Start loading images
				loader.start( 1 );
			} else {
				// Init 3D components
				init3DEngine();
			}
		}

		/**
		 * Shows loader progress info.
		 */
		private function onGalleryLoadingProgress( event:BulkProgressEvent ):void
		{
			if( !_galleryXMLLoaded ) {
				// Show gallery xml loading progress
				trace( 'Loading XML: ' + event.percentLoaded*100 + '%' );
			} else {
				// Show images loading progress
				trace( 'Loading image: ' + event.itemsLoaded + '/' +
					event.itemsTotal + ', ' + event.percentLoaded*100 + '%' );
			}
		}

		/**
		 * Shows loader error info.
		 */
		private function onGalleryLoadingError( event:BulkErrorEvent ):void
		{
			if( !_galleryXMLLoaded ) {
				trace( 'ERROR: Data xml loading failed.' )
			} else {
				trace( 'ERROR: Images loading failed.' )
			}
		}

		/**
		 * Inits 3D engine. Creates viewport, scene, camera and renderer and
		 * also creates and adds scene objects. Creates event listener for the
		 * enter frame event for the scene rendering.
		 */
		private function init3DEngine():void
		{
			// Create viewport
			viewport = new Viewport3D( VIEWPORT_WIDTH, VIEWPORT_HEIGHT, false,
				true );
			addChild( viewport );

			// Create scene
			scene = new Scene3D();

			// Create camera and set its zoom and focus
			camera = new Camera3D();
			camera.zoom = 5;
			camera.focus = 100;

			// Create BasicRenderEngine
			renderer = new BasicRenderEngine();

			// Create Gallery Planes
			createGalleryImages();

			// Add Gallery Planes to the scene
			addGalleryImages();

			// Render the scene every enter frame
			addEventListener( Event.ENTER_FRAME, render );
		}

		/**
		 * Creates display objects from loaded images and updates gallery
		 * objects collection.
		 */
		private function createGalleryImages():void
		{
			galleryPlanesCollection = new Array();
			// Go through all images and create planes
			for( var i:uint=0; i<galleryImagesCollection.length; i++ ) {
				var galleryPlane:DisplayObject3D = createGalleryDisplayObject(
					galleryImagesCollection[i] );
				// Add plane to collection
				galleryPlanesCollection.push( galleryPlane );
			}
		}

		/**
		 * Creates gallery 3D object using loaded image bitmap.
		 *
		 * The method returns DisplayObject3D object, and gallery do not need
		 * to know if there is a plane, cube or sphere, so later I can
		 * rewrite this method to create any objects needed.
		 */
		private function createGalleryDisplayObject( image:String ):DisplayObject3D
		{
			// Get loaded image bitmap using loader getBitmap method
			var bitmap:Bitmap = loader.getBitmap( image );
			var bitmapHolder:Sprite = new Sprite();
			bitmapHolder.addChild( bitmap );
			// Create movie material for the plane
			var material:MovieMaterial = new MovieMaterial( bitmapHolder );
			// Use new material rect property to crop image if it is larger
			// that plane's dimensions
			material.rect = new Rectangle( 0, 0, GALLERY_PLANE_WIDTH,
				GALLERY_PLANE_HEIGHT );
			// Make material double sided and interactive
			material.doubleSided = true;
			material.interactive = true;
			// Create and return plane
			return new Plane( material, GALLERY_PLANE_WIDTH, GALLERY_PLANE_HEIGHT,
				GALLERY_PLANE_SEGMENTS_W, GALLERY_PLANE_SEGMENTS_H );
		}

		/**
		 * Adds gallery display objects to the scene and arranges them. Later I
		 * will add more arrangent strategies to this application, and all I
		 * need to do is to update this method which will get the stategy name
		 * as parameter.
		 */
		private function addGalleryImages():void
		{
			// Create holder for the gallery objects
			_planesHolder = new DisplayObject3D();
			scene.addChild( _planesHolder );

			galleryPlanes = new Array();

			// Add gallery planes
			var flag:int = 1;
			for( var i:uint=0; i<galleryPlanesCollection.length; i++ ) {
				var galleryPlane:Plane = galleryPlanesCollection[i] as Plane;
				galleryPlane.z = _firstPlaneZ + i*200;
				galleryPlane.x = 5000*flag;
				galleryPlanes.push( _planesHolder.addChild( galleryPlane ) );
				flag *= -1;
				Tweener.addTween( galleryPlane, {
					x: 0,
					time:2*(galleryPlanesCollection.length-i)/20+1,
					transition:"easeOutBack"
				} );
			}

			// Listen for the events only from the first gallery object
			galleryPlane = galleryPlanes[_currentPlaneIndex] as Plane;
			galleryPlane.addEventListener(
				InteractiveScene3DEvent.OBJECT_PRESS, onPlanePress );
			galleryPlane.addEventListener(
				InteractiveScene3DEvent.OBJECT_OVER, onPlaneOver );
			galleryPlane.addEventListener(
				InteractiveScene3DEvent.OBJECT_OUT, onPlaneOut );
		}

		/**
		 * Gallery plane mouse press event handler
		 */
		private function onPlanePress( event:InteractiveScene3DEvent ):void
		{
			// Check app state
			if( !_planesAreMoving ) {
				var plane:Plane = event.target as Plane;
				if( !_imageZoomed ) {
					// Change app state
					_imageZoomed = true;

					// Tween clicked plane
					Tweener.addTween( plane, {
						z: -650,
						time:0.5,
						transition:"easeInExpo"
					} );
					Tweener.addTween( plane, {
						z: -600,
						time:0.5,
						delay:0.5,
						transition:"easeOutExpo"
					} );

					// Tween camera to the default position
					Tweener.addTween( _planesHolder, {
						x: 0,
						y: 0,
						rotationY : 0,
						rotationX : 0,
						rotationZ : 0,
						time:0.5,
						transition:"easeOutCubic"
					} );
				} else {
					// Change app state
					_imageZoomed = false;
					_planesAreMoving = true;

					// Remove event listeners from the previous plane
					plane.removeEventListener(
						InteractiveScene3DEvent.OBJECT_PRESS, onPlanePress );
					plane.removeEventListener(
						InteractiveScene3DEvent.OBJECT_OVER, onPlaneOver );
					plane.removeEventListener(
						InteractiveScene3DEvent.OBJECT_OUT, onPlaneOut );

					// Get last plane index and display object
					var lastPlaneIndex:uint = (_currentPlaneIndex-1 < 0 ) ?
						galleryPlanes.length-1 : _currentPlaneIndex-1;
					var lastPlane:Plane = galleryPlanes[ lastPlaneIndex ];

					// Tween every plane one step forward
					var flag:int = 1;
					var planeOrder:uint;
					for( var i:uint=0; i<galleryPlanes.length; i++ ) {
						if( i == _currentPlaneIndex ) continue;
						planeOrder = ( i<_currentPlaneIndex )
							? galleryPlanes.length+i-_currentPlaneIndex
							: i-_currentPlaneIndex;
						var galleryPlane:Plane = galleryPlanesCollection[i]
							as Plane;
						Tweener.addTween( galleryPlane, {
							x: 0,
							y: 0,
							z: galleryPlane.z-200,
							_bezier: {x:1200*flag, y:0, z:galleryPlane.z-100 },
							time: 2*planeOrder/8 + 2,
							transition:"easeInSin"
						} );
						flag *= -1;
					}

					// Tween plane to the end of the row
					Tweener.addTween( plane, {
						x: 0,
						y: 0,
						z: lastPlane.z,
						rotationZ: 0,
						time:2,
						transition:"easeOutCubic",
						onComplete: function():void { _planesAreMoving = false; }
					} );

					// Listner for the current first plane events
					_currentPlaneIndex = (_currentPlaneIndex+1 >
						galleryPlanes.length-1) ? 0 : _currentPlaneIndex+1;
					galleryPlane = galleryPlanes[_currentPlaneIndex] as Plane;
					galleryPlane.addEventListener(
						InteractiveScene3DEvent.OBJECT_PRESS, onPlanePress );
					galleryPlane.addEventListener(
						InteractiveScene3DEvent.OBJECT_OVER, onPlaneOver );
					galleryPlane.addEventListener(
						InteractiveScene3DEvent.OBJECT_OUT, onPlaneOut );
				}
			}
		}

		/**
		 * Gallery plane mouse over event handler
		 */
		private function onPlaneOver( event:InteractiveScene3DEvent ):void
		{
			// Check for the app state
			if( !_imageZoomed && !_planesAreMoving ) {
				// Tween plane
				var plane:Plane = event.target as Plane;
				Tweener.addTween( plane, {
					z: _firstPlaneZ-200,
					time:2,
					transition:"easeInSin"
				} );
			}
		}

		/**
		 * Gallery plane mouse out event handler
		 */
		private function onPlaneOut( event:InteractiveScene3DEvent ):void
		{
			// Check for the app state
			if( !_imageZoomed && !_planesAreMoving ) {
				// Tween plane
				var plane:Plane = event.target as Plane;
				Tweener.addTween( plane, {
					z: _firstPlaneZ,
					time:2,
					transition:"easeInSin"
				} );
			}
		}

		/**
		 * Renders the scene.
		 */
		private function render( event:Event ):void
		{
			// Check for the zoomed image state
			if( !_imageZoomed ) {
				// Calculate distances
				var xDist:Number = root.mouseX - stage.stageWidth * 0.5;
				var yDist:Number = root.mouseY - stage.stageHeight * 0.5;
				var newDist:Number = Math.round(
					Math.sqrt(xDist*xDist + yDist*yDist) ) * 0.08;
				// Calculate camera rotation values
				var cameraRotationX : Number = -xDist * 0.1;
				var cameraRotationY : Number = yDist * 0.1;
				var cameraRotationZ : Number = (xDist*yDist >= 0)
					? -Math.abs(newDist*0.1) : Math.abs(newDist*0.1);
				// Tween camera
				Tweener.addTween( _planesHolder, {
					x: xDist,
					y: yDist,
					rotationY : cameraRotationX,
					rotationX : cameraRotationY,
					rotationZ : cameraRotationZ,
					time:1.2,
					transition:"easeOutCubic"
				} );
			}
			// Render the scene
			renderer.renderScene( scene, camera, viewport );
		}
	}
}